﻿using ELTE.TravelAgency.Admin.Persistence;
using ELTE.TravelAgency.Data;
using ELTE.TravelAgency.Service;
using ELTE.TravelAgency.Service.Models;
using Microsoft.Practices.Unity;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Moq;
using System;
using System.Collections.Generic;
using System.Data.Entity;
using System.Linq;
using System.Web.Http;

namespace TravelAgency.Tests.Service
{
    /// <summary>
    /// Tesztosztály a perzisztencia és a vezérlő integrációjához.
    /// </summary>
    [TestClass]
    public class PersistenceControllerIntegrationTest
    {
        private List<City> _cityData;
        private List<Building> _buildingData;
        private List<BuildingImage> _buildingImageData;
        private Mock<DbSet<City>> _cityMock;
        private Mock<DbSet<Building>> _buildingMock;
        private Mock<DbSet<BuildingImage>> _buildingImageMock;
        private Mock<ITravelAgencyEntities> _entityMock;
        private HttpServer _server;
        private ITravelAgencyPersistence _persistence;

        [TestInitialize]
        public void Initialize()
        {
            // adatok inicializációja
            _cityData = new List<City>
            {
                new City { Id = 1, Name = "TESTCITY" }
            };

            _buildingData = new List<Building>
            {
                new Building { Id = 1, CityId = _cityData[0].Id, City = _cityData[0], Name = "TESTBUILDING1", Features = 0, SeaDistance = 1 },
                new Building { Id = 2, CityId = _cityData[0].Id, City = _cityData[0], Name = "TESTBUILDING2", Features = 1, SeaDistance = 10 },
                new Building { Id = 3, CityId = _cityData[0].Id, City = _cityData[0], Name = "TESTBUILDING3", Features = 2, SeaDistance = 100 }
            };

            _buildingImageData = new List<BuildingImage>
            {
            };

            // entitásmodell inicializációja, amihez le kell generálnunk az adatokhoz tartozó gyűjteményeket (táblákat) is
            // ehhez néhány alap információt be kell állítanunk, lehetővé téve, hogy az adatok lekérdezését biztosítsa a rendszer
            IQueryable<City> queryableCityData = _cityData.AsQueryable();
            _cityMock = new Mock<DbSet<City>>();
            _cityMock.As<IQueryable<City>>().Setup(mock => mock.ElementType).Returns(queryableCityData.ElementType);
            _cityMock.As<IQueryable<City>>().Setup(mock => mock.Expression).Returns(queryableCityData.Expression);
            _cityMock.As<IQueryable<City>>().Setup(mock => mock.Provider).Returns(queryableCityData.Provider);
            _cityMock.As<IQueryable<City>>().Setup(mock => mock.GetEnumerator()).Returns(_cityData.GetEnumerator()); // a korábban megadott listát fogjuk visszaadni

            Building deletedBuilding = null;

            IQueryable<Building> queryableBuildingData = _buildingData.AsQueryable();
            _buildingMock = new Mock<DbSet<Building>>();
            _buildingMock.As<IQueryable<Building>>().Setup(mock => mock.ElementType).Returns(queryableBuildingData.ElementType);
            _buildingMock.As<IQueryable<Building>>().Setup(mock => mock.Expression).Returns(queryableBuildingData.Expression);
            _buildingMock.As<IQueryable<Building>>().Setup(mock => mock.Provider).Returns(queryableBuildingData.Provider);
            _buildingMock.As<IQueryable<Building>>().Setup(mock => mock.GetEnumerator()).Returns(_buildingData.GetEnumerator());

            _buildingMock.Setup(mock => mock.Add(It.IsAny<Building>())).Callback<Building>(building => { _buildingData.Add(building); }).Returns(_buildingData.Last()); // beállítjuk, hogy mi történjen épület hozzáadásakor

            _buildingMock.Setup(mock => mock.Remove(It.IsAny<Building>())).Callback<Building>(building => 
            {
                if (_buildingData.Contains(deletedBuilding))
                    deletedBuilding = building;

                _buildingData.Remove(building);
            }).Returns(deletedBuilding); // beállítjuk, hogy mi történjen épület törlésekor

            _buildingMock.Setup(mock => mock.Include(It.IsAny<String>())).Returns(_buildingMock.Object); // azt is be kell állítanunk, hogy tábla csatolása esetén mi történjen       

            IQueryable<BuildingImage> queryableBuildingImageData = _buildingImageData.AsQueryable();
            _buildingImageMock = new Mock<DbSet<BuildingImage>>();
            _buildingImageMock.As<IQueryable<BuildingImage>>().Setup(mock => mock.ElementType).Returns(queryableBuildingImageData.ElementType);
            _buildingImageMock.As<IQueryable<BuildingImage>>().Setup(mock => mock.Expression).Returns(queryableBuildingImageData.Expression);
            _buildingImageMock.As<IQueryable<BuildingImage>>().Setup(mock => mock.Provider).Returns(queryableBuildingImageData.Provider);
            _buildingImageMock.As<IQueryable<BuildingImage>>().Setup(mock => mock.GetEnumerator()).Returns(_buildingImageData.GetEnumerator());

            // a szimulált entitásmodell csak ezt a két táblát fogja tartalmazni
            _entityMock = new Mock<ITravelAgencyEntities>();
            _entityMock.Setup<DbSet<Building>>(entity => entity.Building).Returns(_buildingMock.Object);
            _entityMock.Setup<DbSet<City>>(entity => entity.City).Returns(_cityMock.Object);
            _entityMock.Setup<DbSet<BuildingImage>>(entity => entity.BuildingImage).Returns(_buildingImageMock.Object);

            // webszolgáltatás inicializációja
            HttpConfiguration config = new HttpConfiguration();
            WebApiConfig.Register(config);

            // IoC tároló az adatbázis kezeléséhez
            UnityContainer container = new UnityContainer();
            container.RegisterInstance<ITravelAgencyEntities>(_entityMock.Object); // itt egy példányt regisztrálunk be

            // az IoC tárolónkat fogja a rendszer használni a függőségek feloldására
            config.DependencyResolver = new UnityResolver(container);

            _server = new HttpServer(config); // memóriában futó HTTP szerver

            _persistence = new TravelAgencyServicePersistence("http://server", _server); // ehhez csatlakozik a kliens
        }

        [TestCleanup]
        public void Cleanup()
        {
            _server.Dispose();
        }
        
        [TestMethod]
        public void PersistenceControllerIntegrationReadBuildingsAsyncTest()
        {
            // meghívjuk a perzisztencián keresztül a szolgáltatást
            List<BuildingDTO> result = _persistence.ReadBuildingsAsync().Result.ToList();

            Assert.AreEqual(_buildingData.Count, result.Count);

            for (Int32 i = 0; i < result.Count; i++)
            {
                Assert.AreEqual(_buildingData[i].Id, result[i].Id);
                Assert.AreEqual(_buildingData[i].Name, result[i].Name);
                Assert.AreEqual(_buildingData[i].City.Id, result[i].City.Id);
                Assert.AreEqual(_buildingData[i].City.Name, result[i].City.Name);
                Assert.AreEqual(_buildingData[i].Comment, result[i].Comment);
                Assert.AreEqual(_buildingData[i].LocationX, result[i].LocationX);
                Assert.AreEqual(_buildingData[i].LocationY, result[i].LocationY);
                Assert.AreEqual(_buildingData[i].SeaDistance, result[i].SeaDistance);
                Assert.AreEqual(_buildingData[i].ShoreId, result[i].ShoreId);
                Assert.AreEqual(_buildingData[i].Features, FeatureDTO.Convert(result[i].Features));
                Assert.AreEqual(_buildingImageData.Count(image => image.BuildingId == _buildingData[i].Id), result[i].Images.Count);
            }
        }

        [TestMethod]
        public void PersistenceControllerIntegrationCreateBuildingsAsyncTest()
        {
            // új épület
            BuildingDTO newBuilding = new BuildingDTO
            {
                Id = 4,
                City = new CityDTO { Id = _cityData[0].Id, Name = _cityData[0].Name },
                Name = "TESTBUILDING4",
                Features = new FeatureDTO[]
                {
                    new FeatureDTO { Id = 0, IsAvailable = true },
                    new FeatureDTO { Id = 1, IsAvailable = false },
                    new FeatureDTO { Id = 2, IsAvailable = true },
                    new FeatureDTO { Id = 3, IsAvailable = false },
                    new FeatureDTO { Id = 4, IsAvailable = true }
                },
                SeaDistance = 1
            };

            Int32 buildingCount = _buildingData.Count;

            // hozzáadjuk az új épületet
            Boolean result = _persistence.CreateBuildingAsync(newBuilding).Result;
            
            // ellenőrizzük, hogy valóban hozzáadott-e az épület
            Assert.AreEqual(buildingCount + 1, _buildingData.Count);
            
            // és az értékek nem változtak
            Assert.AreEqual(_buildingData[_buildingData.Count - 1].Name, newBuilding.Name);
            Assert.AreEqual(_buildingData[_buildingData.Count - 1].CityId, newBuilding.City.Id);
            Assert.AreEqual(_buildingData[_buildingData.Count - 1].Comment, newBuilding.Comment);
            Assert.AreEqual(_buildingData[_buildingData.Count - 1].LocationX, newBuilding.LocationX);
            Assert.AreEqual(_buildingData[_buildingData.Count - 1].LocationY, newBuilding.LocationY);
            Assert.AreEqual(_buildingData[_buildingData.Count - 1].SeaDistance, newBuilding.SeaDistance);
            Assert.AreEqual(_buildingData[_buildingData.Count - 1].ShoreId, newBuilding.ShoreId);
            Assert.AreEqual(_buildingData[_buildingData.Count - 1].Features, FeatureDTO.Convert(newBuilding.Features));
            Assert.AreEqual(_buildingImageData.Count(image => image.BuildingId == _buildingData[_buildingData.Count - 1].Id), newBuilding.Images.Count);
        }

        [TestMethod]
        public void PersistenceControllerIntegrationDeleteBuildingsAsyncTest()
        {
            // törlünk egy létező épületet
            BuildingDTO building = new BuildingDTO { Id = _buildingData[0].Id };

            Int32 buildingCount = _buildingData.Count;

            Boolean result = _persistence.DeleteBuildingAsync(building).Result;

            Assert.IsTrue(result);
            Assert.AreEqual(buildingCount - 1, _buildingData.Count);
            Assert.IsFalse(_buildingData.Any(b => b.Id == building.Id));
 
            building = new BuildingDTO { Id = -1 };

            // törlünk egy nem létező épületet
            result = _persistence.DeleteBuildingAsync(building).Result;

            Assert.IsFalse(result);
        }
    }
}
